from setuptools import setup

setup(
    name='linalgo',
    packages=['linalgo'],
    description=('Python library for Natural Language Processing supporting '
                 'the open annotation standard'),
    version='0.0',
    author='Arnaud Rachez',
    author_email='arnaud@linalgo.com',
    download_url = 'https://github.com/DiscoverSDK/pysimplib/tarball/0.1',
    requires=['numpy', 'scipy'],
)
